/* $Id: protocols.c,v 1.91 2010/09/10 22:55:50 lamonte Exp $
 *
 * protocols.c
 *
 * Additional acknowledgements, news, and updates to the Qodem Project
 * are available at http://qodem.sourceforge.net .
 *
 * This module is licensed under the GNU General Public License Version 2.
 * Please see the file "COPYING" in this directory for more information
 * about the GNU General Public License Version 2.
 *
 *     Copyright (C) 2003  Kevin Lamonte
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 */

/*
 * The download sequence:
 *
 *   STATE = Q_STATE_DOWNLOAD_MENU
 *     protocol_menu_refresh()
 *     protocol_menu_keyboard_handler()
 *   STATE = Q_STATE_DOWNLOAD_PATHDIALOG
 *     protocol_pathdialog_refresh()
 *     protocol_pathdialog_keyboard_handler(): NOP
 *     setup_transfer_program()
 *   STATE = Q_STATE_DOWNLOAD
 *     protocol_transfer_refresh()
 *       state.state = Q_TRANSFER_STATE_INIT
 *     process_transfer_stats()
 *       state.state = Q_TRANSFER_STATE_FILE_INFO
 *       state.state = Q_TRANSFER_STATE_TRANSFER
 *       state.state = Q_TRANSFER_STATE_FILE_DONE
 *     protocol_transfer_keyboard_handler()
 *     close_transfer_program()
 *       state.state = Q_TRANSFER_STATE_END
 *
 * The upload sequence:
 *
 *   STATE = Q_STATE_UPLOAD_MENU
 *     protocol_menu_refresh()
 *     protocol_menu_keyboard_handler()
 *   STATE = Q_STATE_UPLOAD_PATHDIALOG
 *     protocol_pathdialog_refresh()
 *     protocol_pathdialog_keyboard_handler(): NOP
 *     setup_transfer_program()
 *   STATE = Q_STATE_UPLOAD
 *     protocol_transfer_refresh()
 *       state.state = Q_TRANSFER_STATE_INIT
 *     process_transfer_stats()
 *       state.state = Q_TRANSFER_STATE_FILE_INFO
 *       state.state = Q_TRANSFER_STATE_TRANSFER
 *       state.state = Q_TRANSFER_STATE_FILE_DONE
 *     protocol_transfer_keyboard_handler()
 *     close_transfer_program()
 *       state.state = Q_TRANSFER_STATE_END
 *
 * The BATCH upload sequence:
 *
 *   STATE = Q_STATE_UPLOAD_MENU
 *     protocol_menu_refresh()
 *     protocol_menu_keyboard_handler()
 *   STATE = Q_STATE_UPLOAD_PATHDIALOG
 *     protocol_pathdialog_refresh()
 *       STATE = Q_STATE_UPLOAD_BATCH_DIALOG
 *         batch_entry_window()
 *           protocol_pathdialog_refresh()
 *     protocol_pathdialog_keyboard_handler(): NOP
 *   STATE = Q_STATE_UPLOAD_BATCH
 *     setup_transfer_program()
 *   STATE = Q_STATE_UPLOAD_BATCH
 *     protocol_transfer_refresh()
 *       setup_transfer_program()
 *       state.state = Q_TRANSFER_STATE_INIT
 *     process_transfer_stats()
 *       state.state = Q_TRANSFER_STATE_FILE_INFO
 *       state.state = Q_TRANSFER_STATE_TRANSFER
 *       state.state = Q_TRANSFER_STATE_FILE_DONE
 *     protocol_transfer_keyboard_handler()
 *     close_transfer_program()
 *       state.state = Q_TRANSFER_STATE_END
 *   STATE = Q_STATE_CONSOLE
 *
 */
#include "common.h"
#include <ctype.h>
#include <stdlib.h>
#include <errno.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <string.h>
#include <assert.h>
#ifdef __linux
#include <sys/statfs.h>
#endif
#include <sys/statvfs.h>
#include <libgen.h>
#include <stdarg.h>
#include "screen.h"
#include "forms.h"
#include "qodem.h"
#include "console.h"
#include "xmodem.h"
#include "zmodem.h"
#include "kermit.h"
#include "translate.h"
#include "scrollback.h"
#include "states.h"
#include "status.h"
#include "options.h"
#include "phonebook.h"
#include "dialer.h"
#include "codepage.h"
#include "protocols.h"

/* Transfer statistics */
struct q_transfer_stats_struct q_transfer_stats = {
	Q_TRANSFER_STATE_INIT,
	Q_PROTOCOL_ASCII,
	NULL,
	NULL,
	NULL,
	NULL,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};

/* Where transferred files are heading to/from */
char * q_download_location = NULL;

/* List of files for batch upload */
static struct file_info * batch_upload_file_list = NULL;

/* Current file # for batch upload */
static int batch_upload_file_list_i;

/* ------------------------------------------------------------------------
 * ASCII transfer support -------------------------------------------------
 * ------------------------------------------------------------------------
 *
 * Rather than spill to a seperate module, it's so simple I'll just
 * leave it here.
 */

typedef enum {
	ASCII_XFER_STATE_OK,
	ASCII_XFER_STATE_ABORT
} ASCII_XFER_STATE;

typedef enum {
	ASCII_XFER_CRLF_NONE,
	ASCII_XFER_CRLF_STRIP,
	ASCII_XFER_CRLF_ADD
} ASCII_XFER_CRLF_POLICY;

static ASCII_XFER_STATE ascii_xfer_state;
static FILE * ascii_xfer_file = NULL;
static char * ascii_xfer_filename = NULL;
static Q_BOOL ascii_xfer_sending;
static Q_BOOL ascii_xfer_upload_use_xlate_table = Q_FALSE;
static Q_BOOL ascii_xfer_download_use_xlate_table = Q_FALSE;
static ASCII_XFER_CRLF_POLICY ascii_xfer_upload_cr_handling = ASCII_XFER_CRLF_NONE;
static ASCII_XFER_CRLF_POLICY ascii_xfer_upload_lf_handling = ASCII_XFER_CRLF_NONE;
static ASCII_XFER_CRLF_POLICY ascii_xfer_download_cr_handling = ASCII_XFER_CRLF_NONE;
static ASCII_XFER_CRLF_POLICY ascii_xfer_download_lf_handling = ASCII_XFER_CRLF_NONE;

/*
 * Perform CRLF handling on a buffer.
 */
static void ascii_transfer_crlf_handling(const unsigned char * input, const int input_n, unsigned char * output, int * output_n, const ASCII_XFER_CRLF_POLICY cr_policy, const ASCII_XFER_CRLF_POLICY lf_policy) {

	int i, j;

	/* Check if we need to do anything */
	if ((cr_policy == ASCII_XFER_CRLF_NONE) && (lf_policy == ASCII_XFER_CRLF_NONE)) {
		/* Nothing to do */
		memcpy(output, input, input_n);
		*output_n = input_n;
		return;
	}

	/* Iterate through the buffer */
	i = 0;
	j = 0;

	while (i < input_n) {

		if (input[i] == C_CR) {

			/* CR */
			if (cr_policy == ASCII_XFER_CRLF_STRIP) {
				i++;
				continue;
			}
			if (cr_policy == ASCII_XFER_CRLF_ADD) {
				output[j] = input[i];
				j++;
				output[j] = C_LF;
				j++;
				i++;
				continue;
			}
		}

		if (input[i] == C_LF) {

			/* LF */
			if (lf_policy == ASCII_XFER_CRLF_STRIP) {
				i++;
				continue;
			}
			if (lf_policy == ASCII_XFER_CRLF_ADD) {
				output[j] = C_CR;
				j++;
				output[j] = input[i];
				j++;
				i++;
				continue;
			}
		}

		/* Copy the byte over */
		output[j] = input[i];
		j++;
		i++;
	}

	/* Set the output buffer size */
	*output_n = j;
} /* ---------------------------------------------------------------------- */

/*
 * Perform an ASCII file transfer
 */
static void ascii_transfer(unsigned char * input, const int input_n, int * remaining, unsigned char * output, int * output_n, const int output_max) {
	unsigned char working_buffer[2 * Q_BUFFER_SIZE];
	int working_buffer_n;
	char notify_message[DIALOG_MESSAGE_SIZE];
	int rc;

	/* Check my input arguments */
	assert(input_n >= 0);
	assert(input != NULL);
	assert(output != NULL);
	/* assert(*output_n == 0); */
	assert(output_max >= 1);

	if (ascii_xfer_state == ASCII_XFER_STATE_ABORT) {
		return;
	}

	/* Clear working buffer */
	memset(working_buffer, 0, sizeof(working_buffer));
	working_buffer_n = 0;
	int i;

	if (ascii_xfer_sending == Q_TRUE) {
		/* If the outgoing transfer buffer is full, abort */
		if (output_max - *output_n < Q_BUFFER_SIZE) {
			return;
		}

		/*
		 * Send the next few bytes
		 */
		rc = fread(working_buffer, 1, (output_max / 2) - 1, ascii_xfer_file);
		working_buffer_n = rc;

		/* Perform CRLF handling. */
		ascii_transfer_crlf_handling(working_buffer, working_buffer_n, output, output_n, ascii_xfer_upload_cr_handling, ascii_xfer_upload_lf_handling);

		q_transfer_stats.bytes_transfer += *output_n;

		/*
		 * Perform translation table processing
		 */
		if (ascii_xfer_upload_use_xlate_table == Q_TRUE) {
			for (i=0; i < (*output_n); i++) {
				output[i] = q_translate_table_output.map_to[output[i]];
			}
		}

		if (ferror(ascii_xfer_file)) {
			/* Error */
			char notify_message[DIALOG_MESSAGE_SIZE];
			snprintf(notify_message, sizeof(notify_message), _("Error reading from file \"%s\": %s"), ascii_xfer_filename, strerror(errno));
			notify_form(notify_message, 0);

			stop_file_transfer(Q_TRANSFER_STATE_ABORT);
			ascii_xfer_state = ASCII_XFER_STATE_ABORT;
			return;
		}

		if (feof(ascii_xfer_file)) {
			/* End of file */
			stop_file_transfer(Q_TRANSFER_STATE_END);
			time(&q_transfer_stats.end_time);
			q_screen_dirty = Q_TRUE;

			/* Don't do everything again */
			ascii_xfer_state = ASCII_XFER_STATE_ABORT;
		}

	} else {

		/*
		 * Perform translation table processing
		 */
		if (ascii_xfer_download_use_xlate_table == Q_TRUE) {
			for (i=0; i<input_n; i++) {
				input[i] = q_translate_table_input.map_to[input[i]];
			}
		}

		/* Perform CRLF handling. */
		ascii_transfer_crlf_handling(input, input_n, working_buffer, &working_buffer_n, ascii_xfer_download_cr_handling, ascii_xfer_download_lf_handling);

		/*
		 * Save the input bytes to file
		 */
		rc = fwrite(working_buffer, 1, working_buffer_n, ascii_xfer_file);
		if (ferror(ascii_xfer_file)) {
			snprintf(notify_message, sizeof(notify_message), _("Error writing to file \"%s\": %s"), ascii_xfer_filename, strerror(errno));
			notify_form(notify_message, 0);

			stop_file_transfer(Q_TRANSFER_STATE_ABORT);
			ascii_xfer_state = ASCII_XFER_STATE_ABORT;
			return;
		} else {
			/* Flush it */
			fflush(ascii_xfer_file);

			if (rc < working_buffer_n) {
				/* Short file, filesystem is probably full */
				snprintf(notify_message, sizeof(notify_message), _("Error writing to file \"%s\": %s"), ascii_xfer_filename, strerror(errno));
				notify_form(notify_message, 0);

				stop_file_transfer(Q_TRANSFER_STATE_ABORT);
				ascii_xfer_state = ASCII_XFER_STATE_ABORT;
				return;
			}
		}
	}

	/* Run the input side through the console */
	console_process_incoming_data(input, input_n, remaining);

	/* ...and refresh the display */
	q_screen_dirty = Q_TRUE;
	console_refresh(Q_FALSE);
} /* ---------------------------------------------------------------------- */

/*
 * Setup the ASCII protocol for a file transfer
 */
static Q_BOOL ascii_transfer_start(const char * in_filename, const Q_BOOL send) {
	struct stat fstats;

	/* Assume we don't start up successfully */
	ascii_xfer_state = ASCII_XFER_STATE_ABORT;

	/* Pull the options */
	if (strcasecmp(get_option(Q_OPTION_ASCII_UPLOAD_USE_TRANSLATE_TABLE), "true") == 0) {
		ascii_xfer_upload_use_xlate_table = Q_TRUE;
	}
	if (strcasecmp(get_option(Q_OPTION_ASCII_UPLOAD_CR_POLICY), "strip") == 0) {
		ascii_xfer_upload_cr_handling = ASCII_XFER_CRLF_STRIP;
	}
	if (strcasecmp(get_option(Q_OPTION_ASCII_UPLOAD_CR_POLICY), "add") == 0) {
		ascii_xfer_upload_cr_handling = ASCII_XFER_CRLF_ADD;
	}
	if (strcasecmp(get_option(Q_OPTION_ASCII_UPLOAD_LF_POLICY), "strip") == 0) {
		ascii_xfer_upload_lf_handling = ASCII_XFER_CRLF_STRIP;
	}
	if (strcasecmp(get_option(Q_OPTION_ASCII_UPLOAD_LF_POLICY), "add") == 0) {
		ascii_xfer_upload_lf_handling = ASCII_XFER_CRLF_ADD;
	}
	if (strcasecmp(get_option(Q_OPTION_ASCII_DOWNLOAD_USE_TRANSLATE_TABLE), "true") == 0) {
		ascii_xfer_download_use_xlate_table = Q_TRUE;
	}
	if (strcasecmp(get_option(Q_OPTION_ASCII_DOWNLOAD_CR_POLICY), "strip") == 0) {
		ascii_xfer_download_cr_handling = ASCII_XFER_CRLF_STRIP;
	}
	if (strcasecmp(get_option(Q_OPTION_ASCII_DOWNLOAD_CR_POLICY), "add") == 0) {
		ascii_xfer_download_cr_handling = ASCII_XFER_CRLF_ADD;
	}
	if (strcasecmp(get_option(Q_OPTION_ASCII_DOWNLOAD_LF_POLICY), "strip") == 0) {
		ascii_xfer_download_lf_handling = ASCII_XFER_CRLF_STRIP;
	}
	if (strcasecmp(get_option(Q_OPTION_ASCII_DOWNLOAD_LF_POLICY), "add") == 0) {
		ascii_xfer_download_lf_handling = ASCII_XFER_CRLF_ADD;
	}


	if (send == Q_TRUE) {
		/* Pull the file size */
		if (lstat(in_filename, &fstats) < 0) {
			return Q_FALSE;
		}

		if ((ascii_xfer_file = fopen(in_filename, "rb")) == NULL) {
			return Q_FALSE;
		}

		q_transfer_stats.bytes_total = fstats.st_size;
		q_transfer_stats.block_size = 128;
		q_transfer_stats.blocks = fstats.st_size / 128;
		if ((q_transfer_stats.blocks % 128) > 0) {
			q_transfer_stats.blocks++;
		}

	} else {
		if ((ascii_xfer_file = fopen(in_filename, "w+b")) == NULL) {
			return Q_FALSE;
		}
	}

	ascii_xfer_filename = Xstrdup(in_filename, __FILE__, __LINE__);
	ascii_xfer_sending = send;

	/* Let's go! */
	ascii_xfer_state = ASCII_XFER_STATE_OK;

	return Q_TRUE;
} /* ---------------------------------------------------------------------- */

/*
 * End an ASCII transfer
 */
static void ascii_transfer_stop(const Q_BOOL save_partial) {

	if ((save_partial == Q_TRUE) || (ascii_xfer_sending == Q_TRUE)) {
		if (ascii_xfer_file != NULL) {
			fflush(ascii_xfer_file);
			fclose(ascii_xfer_file);
		}
	} else {
		if (ascii_xfer_file != NULL) {
			fclose(ascii_xfer_file);
			unlink(ascii_xfer_filename);
		}
	}
	ascii_xfer_file = NULL;
	if (ascii_xfer_filename != NULL) {
		Xfree(ascii_xfer_filename, __FILE__, __LINE__);
	}
	ascii_xfer_filename = NULL;
} /* ---------------------------------------------------------------------- */

/* ------------------------------------------------------------------------
 * ASCII transfer support -------------------------------------------------
 * ------------------------------------------------------------------------ */

/*
 * Convenience functions to handle the strings inside q_transfer_stats_struct.
 * To avoid memory leaks I Xstrdup() them in and then Xfree() them out when
 * I switch to Q_STATE_CONSOLE.
 */
void set_transfer_stats_protocol_name(const char * new_string) {
	/*
	 * Do nothing if being set to what it already is.
	 * Note this is a pointer comparison, not a string
	 * comparison.
	 */
	if (new_string == q_transfer_stats.protocol_name) {
		return;
	}

	if (q_transfer_stats.protocol_name != NULL) {
		Xfree(q_transfer_stats.protocol_name, __FILE__, __LINE__);
	}
	q_transfer_stats.protocol_name = Xstrdup(new_string, __FILE__, __LINE__);
} /* ---------------------------------------------------------------------- */

void set_transfer_stats_filename(const char * new_string) {
	/*
	 * Do nothing if being set to what it already is.
	 * Note this is a pointer comparison, not a string
	 * comparison.
	 */
	if (new_string == q_transfer_stats.filename) {
		return;
	}

	if (q_transfer_stats.filename != NULL) {
		Xfree(q_transfer_stats.filename, __FILE__, __LINE__);
	}
	q_transfer_stats.filename = Xstrdup(new_string, __FILE__, __LINE__);
} /* ---------------------------------------------------------------------- */

void set_transfer_stats_pathname(const char * new_string) {
	/*
	 * Do nothing if being set to what it already is.
	 * Note this is a pointer comparison, not a string
	 * comparison.
	 */
	if (new_string == q_transfer_stats.pathname) {
		return;
	}

	if (q_transfer_stats.pathname != NULL) {
		Xfree(q_transfer_stats.pathname, __FILE__, __LINE__);
	}
	q_transfer_stats.pathname = Xstrdup(new_string, __FILE__, __LINE__);
} /* ---------------------------------------------------------------------- */

void set_transfer_stats_last_message(const char * format, ...) {
	/*
	 * Since I always generate the string here I don't
	 * need to check that I'm about to Xfree() myself.
	 */
	char outbuf[DIALOG_MESSAGE_SIZE];
	va_list arglist;
	memset(outbuf, 0, sizeof(outbuf));
	va_start(arglist, format);
	vsprintf((char *)(outbuf+strlen(outbuf)), format, arglist);
	va_end(arglist);

	if (q_transfer_stats.last_message != NULL) {
		Xfree(q_transfer_stats.last_message, __FILE__, __LINE__);
	}
	q_transfer_stats.last_message = Xstrdup(outbuf, __FILE__, __LINE__);

	/* Report the message immediately */
	q_screen_dirty = Q_TRUE;
} /* ---------------------------------------------------------------------- */

/*
 * Process incoming data within a file transfer
 */
void protocol_process_data(unsigned char * input, const int input_n, int * remaining, unsigned char * output, int * output_n, const int output_max) {

	if ((q_transfer_stats.state == Q_TRANSFER_STATE_ABORT) || (q_transfer_stats.state == Q_TRANSFER_STATE_END)) {
		return;
	}

	switch (q_transfer_stats.protocol) {

	case Q_PROTOCOL_ASCII:
		ascii_transfer(input, input_n, remaining, output, output_n, output_max);
		break;
	case Q_PROTOCOL_KERMIT:
		/*
		 * Kermit - I don't need to pass remaining because
		 * kermit.c does its own buffering in packet_buffer .
		 */
		kermit(input, input_n, output, output_n, output_max);
		/* Kermit always disposes of everything sent to it. */
		*remaining = 0;
		break;
	case Q_PROTOCOL_XMODEM:
	case Q_PROTOCOL_XMODEM_CRC:
	case Q_PROTOCOL_XMODEM_RELAXED:
	case Q_PROTOCOL_XMODEM_1K:
	case Q_PROTOCOL_YMODEM:
	case Q_PROTOCOL_XMODEM_1K_G:
	case Q_PROTOCOL_YMODEM_G:
		/*
		 * Yes, ALL of these protocols use the same function to
		 * transmit data.
		 */
		xmodem(input, input_n, remaining, output, output_n, output_max);
		break;
	case Q_PROTOCOL_ZMODEM:
		/*
		 * Zmodem - I don't need to pass remaining because
		 * zmodem.c does its own buffering in packet_buffer .
		 */
		zmodem(input, input_n, output, output_n, output_max);
		/* Zmodem always disposes of everything sent to it. */
		*remaining = 0;
		break;
	default:
		/* Should never get here */
		assert(1 == 0);
	}

} /* ---------------------------------------------------------------------- */

/*
 * clear_stats - reset transfer statistics
 */
static void clear_stats() {
	int i;

	/* Setup stats */
	if (q_transfer_stats.filename != NULL) {
		Xfree(q_transfer_stats.filename, __FILE__, __LINE__);
		q_transfer_stats.filename = NULL;
	}
	if (q_transfer_stats.pathname != NULL) {
		Xfree(q_transfer_stats.pathname, __FILE__, __LINE__);
		q_transfer_stats.pathname = NULL;
	}
	if (q_transfer_stats.protocol_name != NULL) {
		Xfree(q_transfer_stats.protocol_name, __FILE__, __LINE__);
		q_transfer_stats.protocol_name = NULL;
	}
	if (q_transfer_stats.last_message != NULL) {
		Xfree(q_transfer_stats.last_message, __FILE__, __LINE__);
		q_transfer_stats.last_message = NULL;
	}
	q_transfer_stats.state			= Q_TRANSFER_STATE_INIT;
	q_transfer_stats.bytes_total		= 0;
	q_transfer_stats.bytes_transfer		= 0;
	q_transfer_stats.error_count		= 0;
	q_transfer_stats.blocks			= 0;
	q_transfer_stats.block_size		= 0;
	q_transfer_stats.blocks_transfer	= 0;
	q_transfer_stats.batch_bytes_total	= 0;
	q_transfer_stats.batch_bytes_transfer	= 0;

	/* Batch uploads: calculate q_transfer_stats.batch_bytes_total */
	if (q_program_state == Q_STATE_UPLOAD_BATCH) {
		for (i=0;; i++) {
			if (batch_upload_file_list[i].name == NULL) {
				break;
			}
			q_transfer_stats.batch_bytes_total += batch_upload_file_list[i].fstats.st_size;
		}
	}
	if (q_program_state == Q_STATE_UPLOAD) {
		struct stat fstats;
		if (q_transfer_stats.filename != NULL) {
			if (lstat(q_transfer_stats.filename, &fstats) < 0) {
				fprintf(stderr, "Can't stat %s: %s\n", q_transfer_stats.filename, strerror(errno));
			}
			q_transfer_stats.bytes_total = fstats.st_size;
		}
	}

} /* ---------------------------------------------------------------------- */

/*
 * start_file_transfer - Transfer a file between q_child_tty_fd and q_download_location.
 *                 For ASCII and Xmodem, q_download_location is a full filename.
 *                 For Kermit, Ymodem, and Zmodem, q_download_location is a path name.
 */
void start_file_transfer() {
	char * basename_arg;
	char * dirname_arg;

	/* Refresh the background */
	q_screen_dirty = Q_TRUE;
	console_refresh(Q_FALSE);

	/* Clear stats */
	clear_stats();

	/* Set protocol name */
	switch (q_transfer_stats.protocol) {

	case Q_PROTOCOL_ASCII:
		set_transfer_stats_protocol_name(_("ASCII"));
		break;
	case Q_PROTOCOL_KERMIT:
		set_transfer_stats_protocol_name(_("Kermit"));
		break;
	case Q_PROTOCOL_XMODEM:
		set_transfer_stats_protocol_name(_("Xmodem"));
		break;
	case Q_PROTOCOL_XMODEM_CRC:
		set_transfer_stats_protocol_name(_("Xmodem CRC"));
		break;
	case Q_PROTOCOL_XMODEM_RELAXED:
		set_transfer_stats_protocol_name(_("Xmodem Relaxed"));
		break;
	case Q_PROTOCOL_XMODEM_1K:
		set_transfer_stats_protocol_name(_("XMODEM_1K"));
		break;
	case Q_PROTOCOL_XMODEM_1K_G:
		set_transfer_stats_protocol_name(_("Xmodem-1K/G"));
		break;
	case Q_PROTOCOL_YMODEM:
		set_transfer_stats_protocol_name(_("Ymodem Batch"));
		break;
	case Q_PROTOCOL_YMODEM_G:
		set_transfer_stats_protocol_name(_("Ymodem/G Batch"));
		break;
	case Q_PROTOCOL_ZMODEM:
		set_transfer_stats_protocol_name(_("Zmodem Batch"));
		break;
	default:
		/* Should never get here */
		assert(1 == 0);
	}

	/* Log it */
	switch (q_transfer_stats.protocol) {

	case Q_PROTOCOL_ASCII:
	case Q_PROTOCOL_XMODEM:
	case Q_PROTOCOL_XMODEM_CRC:
	case Q_PROTOCOL_XMODEM_RELAXED:
	case Q_PROTOCOL_XMODEM_1K:
	case Q_PROTOCOL_XMODEM_1K_G:
		/* Single-file protocols */
		if (q_program_state == Q_STATE_DOWNLOAD) {
			qlog(_("DOWNLOAD BEGIN: protocol %s, filename %s\n"), q_transfer_stats.protocol_name, q_download_location);
		} else {
			qlog(_("UPLOAD BEGIN: protocol %s, filename %s\n"), q_transfer_stats.protocol_name, q_download_location);
		}
		break;

	default:
		/* Batch protocols */
		if (q_program_state == Q_STATE_DOWNLOAD) {
			qlog(_("DOWNLOAD BEGIN: protocol %s\n"), q_transfer_stats.protocol_name);
		} else {
			qlog(_("UPLOAD BEGIN: protocol %s\n"), q_transfer_stats.protocol_name);
		}
		break;
	}

	/* Cursor off */
	q_cursor_off();

	if (q_program_state != Q_STATE_UPLOAD_BATCH) {
		/* Strip trailing '/' from q_download_location */
		while (q_download_location[strlen(q_download_location)-1] == '/') {
			q_download_location[strlen(q_download_location)-1] = '\0';
		}
	}

	/* Start the protocols */
	switch (q_transfer_stats.protocol) {

	case Q_PROTOCOL_ASCII:
		if (ascii_transfer_start(q_download_location, ((q_program_state == Q_STATE_UPLOAD) ? Q_TRUE : Q_FALSE)) == Q_FALSE) {
			/* Couldn't start the protocol, switch back to console. */
			switch_state(Q_STATE_CONSOLE);
			return;
		}
		break;
	case Q_PROTOCOL_KERMIT:
		if (kermit_start(batch_upload_file_list, q_download_location, ((q_program_state == Q_STATE_DOWNLOAD) ? Q_FALSE : Q_TRUE)) == Q_FALSE) {
			/* Couldn't start the protocol, switch back to console. */
			switch_state(Q_STATE_CONSOLE);
			return;
		}
		break;
	case Q_PROTOCOL_XMODEM:
		if (xmodem_start(q_download_location, ((q_program_state == Q_STATE_UPLOAD) ? Q_TRUE : Q_FALSE), X_NORMAL) == Q_FALSE) {
			/* Couldn't start the protocol, switch back to console. */
			switch_state(Q_STATE_CONSOLE);
			return;
		}
		break;
	case Q_PROTOCOL_XMODEM_CRC:
		if (xmodem_start(q_download_location, ((q_program_state == Q_STATE_UPLOAD) ? Q_TRUE : Q_FALSE), X_CRC) == Q_FALSE) {
			/* Couldn't start the protocol, switch back to console. */
			switch_state(Q_STATE_CONSOLE);
			return;
		}
		break;
	case Q_PROTOCOL_XMODEM_RELAXED:
		if (xmodem_start(q_download_location, ((q_program_state == Q_STATE_UPLOAD) ? Q_TRUE : Q_FALSE), X_RELAXED) == Q_FALSE) {
			/* Couldn't start the protocol, switch back to console. */
			switch_state(Q_STATE_CONSOLE);
			return;
		}
		break;
	case Q_PROTOCOL_XMODEM_1K:
		if (xmodem_start(q_download_location, ((q_program_state == Q_STATE_UPLOAD) ? Q_TRUE : Q_FALSE), X_1K) == Q_FALSE) {
			/* Couldn't start the protocol, switch back to console. */
			switch_state(Q_STATE_CONSOLE);
			return;
		}
		break;
	case Q_PROTOCOL_XMODEM_1K_G:
		if (xmodem_start(q_download_location, ((q_program_state == Q_STATE_UPLOAD) ? Q_TRUE : Q_FALSE), X_1K_G) == Q_FALSE) {
			/* Couldn't start the protocol, switch back to console. */
			switch_state(Q_STATE_CONSOLE);
			return;
		}
		break;
	case Q_PROTOCOL_YMODEM:
		if (ymodem_start(batch_upload_file_list, q_download_location, ((q_program_state == Q_STATE_UPLOAD_BATCH) ? Q_TRUE : Q_FALSE), Y_NORMAL) == Q_FALSE) {
			/* Couldn't start the protocol, switch back to console. */
			switch_state(Q_STATE_CONSOLE);
			return;
		}
		break;
	case Q_PROTOCOL_YMODEM_G:
		if (ymodem_start(batch_upload_file_list, q_download_location, ((q_program_state == Q_STATE_DOWNLOAD) ? Q_FALSE : Q_TRUE), Y_G) == Q_FALSE) {
			/* Couldn't start the protocol, switch back to console. */
			switch_state(Q_STATE_CONSOLE);
			return;
		}
		break;
	case Q_PROTOCOL_ZMODEM:
		/* By default, always use CRC32 */
		if (zmodem_start(batch_upload_file_list, q_download_location, ((q_program_state == Q_STATE_DOWNLOAD) ? Q_FALSE : Q_TRUE), Z_CRC32) == Q_FALSE) {
			/* Couldn't start the protocol, switch back to console. */
			switch_state(Q_STATE_CONSOLE);
			return;
		}
		break;
	default:
		/* Should never get here */
		assert(1 == 0);
	}

	/* Setup filename and basename */
	switch (q_transfer_stats.protocol) {

	case Q_PROTOCOL_KERMIT:
		/* Filename is setup by kermit_start() or during transfer */
		break;
	case Q_PROTOCOL_ASCII:
	case Q_PROTOCOL_XMODEM:
	case Q_PROTOCOL_XMODEM_CRC:
	case Q_PROTOCOL_XMODEM_RELAXED:
	case Q_PROTOCOL_XMODEM_1K:
	case Q_PROTOCOL_XMODEM_1K_G:
		/* Note that basename and dirname modify the arguments */
		basename_arg = Xstrdup(q_download_location, __FILE__, __LINE__);
		dirname_arg = Xstrdup(q_download_location, __FILE__, __LINE__);
		set_transfer_stats_filename(basename(basename_arg));
		set_transfer_stats_pathname(dirname(dirname_arg));
		/* Free the copies passed to basename() and dirname() */
		Xfree(basename_arg, __FILE__, __LINE__);
		Xfree(dirname_arg, __FILE__, __LINE__);
		break;

	case Q_PROTOCOL_YMODEM:
	case Q_PROTOCOL_YMODEM_G:
		/* Filename is setup by ymodem_start() or during transfer */
		break;
	case Q_PROTOCOL_ZMODEM:
		/* Filename is setup by zmodem_start() or during transfer */
		break;

	default:
		/* Should never get here */
		assert(1 == 0);
	}

	/* Record the time */
	time(&q_transfer_stats.file_start_time);
	time(&q_transfer_stats.batch_start_time);

} /* ---------------------------------------------------------------------- */

/*
 * stop_file_transfer - Stop a running transfer
 */
void stop_file_transfer(const Q_TRANSFER_STATE new_state) {
	int i;

	switch (q_transfer_stats.protocol) {

	case Q_PROTOCOL_ASCII:
		ascii_transfer_stop(Q_TRUE);
		break;
	case Q_PROTOCOL_KERMIT:
		kermit_stop(Q_TRUE);
		break;
	case Q_PROTOCOL_XMODEM:
	case Q_PROTOCOL_XMODEM_CRC:
	case Q_PROTOCOL_XMODEM_RELAXED:
	case Q_PROTOCOL_XMODEM_1K:
	case Q_PROTOCOL_XMODEM_1K_G:
		xmodem_stop(Q_TRUE);
		break;
	case Q_PROTOCOL_YMODEM:
	case Q_PROTOCOL_YMODEM_G:
		ymodem_stop(Q_TRUE);
		break;
	case Q_PROTOCOL_ZMODEM:
		zmodem_stop(Q_TRUE);
		break;
	default:
		/* Should never get here */
		assert(1 == 0);

	}

	q_transfer_stats.state = new_state;
	q_screen_dirty = Q_TRUE;
	time(&q_transfer_stats.end_time);

	/* Free the batch upload list */
	if (batch_upload_file_list != NULL) {
		for (i = 0; batch_upload_file_list[i].name != NULL; i++) {
			Xfree(batch_upload_file_list[i].name, __FILE__, __LINE__);
		}
		Xfree(batch_upload_file_list, __FILE__, __LINE__);
		batch_upload_file_list_i = 0;
		batch_upload_file_list = NULL;
	}

	if (q_download_location != NULL) {
		Xfree(q_download_location, __FILE__, __LINE__);
		q_download_location = NULL;
	}

	/* Log it */
	switch (q_transfer_stats.protocol) {

	case Q_PROTOCOL_ASCII:
	case Q_PROTOCOL_XMODEM:
	case Q_PROTOCOL_XMODEM_CRC:
	case Q_PROTOCOL_XMODEM_RELAXED:
	case Q_PROTOCOL_XMODEM_1K:
	case Q_PROTOCOL_XMODEM_1K_G:
		/* Single-file protocols */
		if (q_program_state == Q_STATE_DOWNLOAD) {
			if (new_state == Q_TRANSFER_STATE_ABORT) {
				qlog(_("DOWNLOAD ABORTED: protocol %s, filename %s\n"), q_transfer_stats.protocol_name);
			} else {
				qlog(_("DOWNLOAD FILE COMPLETE: protocol %s, filename %s, filesize %d\n"), q_transfer_stats.protocol_name, q_download_location, q_transfer_stats.bytes_total);
			}
		} else {
			if (new_state == Q_TRANSFER_STATE_ABORT) {
				qlog(_("UPLOAD ABORTED: protocol %s, filename %s\n"), q_transfer_stats.protocol_name);
			} else {
				qlog(_("UPLOAD FILE COMPLETE: protocol %s, filename %s, filesize %d\n"), q_transfer_stats.protocol_name, q_download_location, q_transfer_stats.bytes_total);
			}
		}
		break;

	default:
		/* Batch protocols */
		if (q_program_state == Q_STATE_DOWNLOAD) {
			if (new_state == Q_TRANSFER_STATE_ABORT) {
				qlog(_("DOWNLOAD ABORTED: protocol %s\n"), q_transfer_stats.protocol_name);
			} else {
				qlog(_("DOWNLOAD END: protocol %s\n"), q_transfer_stats.protocol_name, q_download_location, q_transfer_stats.bytes_total);
			}
		} else {
			if (new_state == Q_TRANSFER_STATE_ABORT) {
				qlog(_("UPLOAD ABORTED: protocol %s\n"), q_transfer_stats.protocol_name);
			} else {
				qlog(_("UPLOAD END: protocol %s\n"), q_transfer_stats.protocol_name, q_download_location, q_transfer_stats.bytes_total);
			}
		}
		break;
	}

} /* ---------------------------------------------------------------------- */

/*
 * protocol_menu_refresh
 */
void protocol_menu_refresh() {
	char * status_string;
	/* Number of digits in the hard drive size display.  10^32 is pretty darn big. */
	char size_string[32];
	int status_left_stop;
	char * message;
	int message_left;
	int window_left;
	int window_top;
	int window_height = 13;
	int window_length;
	int i;
	struct statvfs buf;
	unsigned long free_kbytes = 0;

	if (q_screen_dirty == Q_FALSE) {
		return;
	}

	if (q_program_state == Q_STATE_DOWNLOAD_MENU) {
		window_height++;
		message = _("Download Protocols");
	} else {
		message = _("Upload Protocols");
	}

	/* Put up the status line */
	screen_put_color_hline_yx(HEIGHT - 1, 0, cp437_chars[HATCH], WIDTH, Q_COLOR_STATUS);

	status_string = _(" LETTER-Select a Protocol for the File Transfer   ESC/`-Exit ");
	status_left_stop = WIDTH - strlen(status_string);
	if (status_left_stop <= 0) {
		status_left_stop = 0;
	} else {
		status_left_stop /= 2;
	}
	screen_put_color_str_yx(HEIGHT - 1, status_left_stop, status_string, Q_COLOR_STATUS);

	memset(size_string, 0, sizeof(size_string));
	if (q_program_state == Q_STATE_DOWNLOAD_MENU) {
		statvfs(get_option(Q_OPTION_DOWNLOAD_DIR), &buf);
		free_kbytes = buf.f_bavail * buf.f_bsize / 1024;
		sprintf(size_string, _("Free Space  %'-lu k"), free_kbytes);
	}

	if (strlen(message) > strlen(size_string)) {
		window_length = strlen(message);
	} else {
		window_length = strlen(size_string);
	}

	/* Add room for border + 1 spaces on each side */
	window_length += 4;

	if (q_program_state == Q_STATE_UPLOAD_MENU) {
		/* Upload window needs a tad more space */
		window_length += 2;
	}

	/* Window will be centered on the screen */
	window_left = WIDTH - 1 - window_length;
	if (window_left < 0) {
		window_left = 0;
	} else {
		window_left /= 2;
	}
	window_top = HEIGHT - 1 - window_height;
	if (window_top < 0) {
		window_top = 0;
	} else {
		window_top /= 10;
	}

	/* Draw the sub-window */
	screen_draw_box(window_left, window_top, window_left + window_length, window_top + window_height);

	/* Place the title */
	message_left = window_length - (strlen(message) + 2);
	if (message_left < 0) {
		message_left = 0;
	} else {
		message_left /= 2;
	}
	screen_put_color_printf_yx(window_top + 0, window_left + message_left, Q_COLOR_WINDOW_BORDER, " %s ", message);

	i = 1;
	if (q_program_state == Q_STATE_DOWNLOAD_MENU) {
		screen_put_color_str_yx(window_top + 1, window_left + 2, _("Free Space"), Q_COLOR_MENU_COMMAND);
		screen_put_color_printf(Q_COLOR_MENU_TEXT, "  %'-lu k", free_kbytes);
		i++;
	}
	screen_put_color_str_yx(window_top + i, window_left + 2, "A", Q_COLOR_MENU_COMMAND);
	screen_put_color_str(_(" - Ascii"), Q_COLOR_MENU_TEXT);
	i++;
	screen_put_color_str_yx(window_top + i, window_left + 2, "K", Q_COLOR_MENU_COMMAND);
	screen_put_color_str(_(" - Kermit"), Q_COLOR_MENU_TEXT);
	i++;
	screen_put_color_str_yx(window_top + i, window_left + 2, "X", Q_COLOR_MENU_COMMAND);
	screen_put_color_str(_(" - Xmodem"), Q_COLOR_MENU_TEXT);
	i++;
	screen_put_color_str_yx(window_top + i, window_left + 2, "C", Q_COLOR_MENU_COMMAND);
	screen_put_color_str(_(" - Xmodem CRC"), Q_COLOR_MENU_TEXT);
	i++;
	screen_put_color_str_yx(window_top + i, window_left + 2, "R", Q_COLOR_MENU_COMMAND);
	screen_put_color_str(_(" - Xmodem Relaxed"), Q_COLOR_MENU_TEXT);
	i++;
	screen_put_color_str_yx(window_top + i, window_left + 2, "O", Q_COLOR_MENU_COMMAND);
	screen_put_color_str(_(" - Xmodem-1K"), Q_COLOR_MENU_TEXT);
	i++;
	screen_put_color_str_yx(window_top + i, window_left + 2, "Y", Q_COLOR_MENU_COMMAND);
	screen_put_color_str(_(" - Ymodem Batch"), Q_COLOR_MENU_TEXT);
	i++;
	screen_put_color_str_yx(window_top + i, window_left + 2, "Z", Q_COLOR_MENU_COMMAND);
	screen_put_color_str(_(" - Zmodem Batch"), Q_COLOR_MENU_TEXT);
	i++;
	screen_put_color_str_yx(window_top + i, window_left + 2, "F", Q_COLOR_MENU_COMMAND);
	screen_put_color_str(_(" - Xmodem-1K/G"), Q_COLOR_MENU_TEXT);
	i++;
	screen_put_color_str_yx(window_top + i, window_left + 2, "G", Q_COLOR_MENU_COMMAND);
	screen_put_color_str(_(" - Ymodem/G Batch"), Q_COLOR_MENU_TEXT);
	i++;

	/* Prompt */
	screen_put_color_str_yx(window_top + i, window_left + 2, _("Your Choice ? "), Q_COLOR_MENU_COMMAND);

	screen_flush();
	q_screen_dirty = Q_FALSE;
} /* ---------------------------------------------------------------------- */

/*
 * protocol_menu_keyboard_handler
 */
void protocol_menu_keyboard_handler(const int keystroke, const int flags) {

	switch (keystroke) {

	case 'a':
	case 'A':
		q_transfer_stats.protocol = Q_PROTOCOL_ASCII;
		break;
	case 'k':
	case 'K':
		q_transfer_stats.protocol = Q_PROTOCOL_KERMIT;
		break;
	case 'x':
	case 'X':
		q_transfer_stats.protocol = Q_PROTOCOL_XMODEM;
		break;
	case 'c':
	case 'C':
		q_transfer_stats.protocol = Q_PROTOCOL_XMODEM_CRC;
		break;
	case 'r':
	case 'R':
		q_transfer_stats.protocol = Q_PROTOCOL_XMODEM_RELAXED;
		break;
	case 'o':
	case 'O':
		q_transfer_stats.protocol = Q_PROTOCOL_XMODEM_1K;
		break;
	case 'y':
	case 'Y':
		q_transfer_stats.protocol = Q_PROTOCOL_YMODEM;
		break;
	case 'z':
	case 'Z':
		q_transfer_stats.protocol = Q_PROTOCOL_ZMODEM;
		break;
	case 'f':
	case 'F':
		q_transfer_stats.protocol = Q_PROTOCOL_XMODEM_1K_G;
		break;
	case 'g':
	case 'G':
		q_transfer_stats.protocol = Q_PROTOCOL_YMODEM_G;
		break;
	case '`':
		/* Backtick works too */
	case KEY_ESCAPE:
		/* ESC return to TERMINAL mode */
		switch_state(Q_STATE_CONSOLE);

		/* The ABORT exit point */
		return;

	default:
		/* Ignore keystroke */
		return;

	}

	/* Protocol selected, switch to prompt. */
	if (q_program_state == Q_STATE_DOWNLOAD_MENU) {
		switch_state(Q_STATE_DOWNLOAD_PATHDIALOG);
	} else {
		switch_state(Q_STATE_UPLOAD_PATHDIALOG);
	}

	/* The OK exit point */

} /* ---------------------------------------------------------------------- */

/*
 * protocol_pathdialog_refresh - This screen handler is unique in that it blocks
 *                               the main loop.
 */
void protocol_pathdialog_refresh() {
	/* Refresh the background */
	q_screen_dirty = Q_TRUE;
	console_refresh(Q_FALSE);

	/* Special case: I could be called from inside batch_entry_window() */
	if (q_program_state == Q_STATE_UPLOAD_BATCH_DIALOG) {
		return;
	}

	if (q_download_location != NULL) {
		Xfree(q_download_location, __FILE__, __LINE__);
		q_download_location = NULL;
	}

	switch (q_transfer_stats.protocol) {

	case Q_PROTOCOL_KERMIT:
	case Q_PROTOCOL_YMODEM:
	case Q_PROTOCOL_YMODEM_G:
	case Q_PROTOCOL_ZMODEM:
		/*
		 * Kermit, Ymodem and Zmodem get the filename from the download itself,
		 * so prompt for a pathname rather than a filename.
		 */
		if (q_program_state == Q_STATE_DOWNLOAD_PATHDIALOG) {
			q_download_location = save_form(_("Download Directory"), get_option(Q_OPTION_DOWNLOAD_DIR), Q_TRUE, Q_FALSE);
		} else {
			/* Special case: batch entry window */
			switch_state(Q_STATE_UPLOAD_BATCH_DIALOG);
			batch_upload_file_list = batch_entry_window(get_option(Q_OPTION_UPLOAD_DIR), Q_TRUE);
			if (batch_upload_file_list != NULL) {
				/* Begin uploading */
				batch_upload_file_list_i = 0;
				switch_state(Q_STATE_UPLOAD_BATCH);
				start_file_transfer();

			} else {
				/* Abort */
				switch_state(Q_STATE_CONSOLE);
			}

			/*
			 * I could let this fall out below, but that would be
			 * slightly confusing. This function is zig-zaggy enough
			 * already.
			 */
			return;
		}
		break;

	default:
		if (q_program_state == Q_STATE_DOWNLOAD_PATHDIALOG) {
			q_download_location = save_form(_("Download File"), get_option(Q_OPTION_DOWNLOAD_DIR), Q_FALSE, Q_TRUE);
		} else {
			q_download_location = save_form(_("Upload File"), get_option(Q_OPTION_UPLOAD_DIR), Q_FALSE, Q_FALSE);
		}
		break;

	}

	if (q_program_state == Q_STATE_DOWNLOAD_PATHDIALOG) {
		/* Begin downloading */
		switch_state(Q_STATE_DOWNLOAD);
	} else {
		/* Begin uploading */
		switch_state(Q_STATE_UPLOAD);
	}

	/* Start the transfer */
	if (q_download_location != NULL) {
		start_file_transfer();
	} else {
		/* Abort */
		switch_state(Q_STATE_CONSOLE);
	}

} /* ---------------------------------------------------------------------- */

/*
 * protocol_pathdialog_keyboard_handler
 */
void protocol_pathdialog_keyboard_handler(const int keystroke, const int flags) {
	/* This function gets called after protocol_pathdialog_refresh(),
	   so it needs to exist, but it doesn't actually do anything.
	   protocol_pathdialog_refresh() calls functions in forms.c
	   that do their own internal keyboard handling. */

	/* NOP */
	return;
} /* ---------------------------------------------------------------------- */

/*
 * shorten_string - Reduce string to size length.
 */
static void shorten_string(char * string, const int length) {
	if (string == NULL) {
		return;
	}

	if (strlen(string) < length - 4) {
		return;
	}

	string[length] = '\0';
	string[length-1] = '.';
	string[length-2] = '.';
	string[length-3] = '.';

} /* ---------------------------------------------------------------------- */

/*
 * protocol_transfer_keyboard_handler
 */
void protocol_transfer_keyboard_handler(const int keystroke, const int flags) {
	switch (keystroke) {
	case 's':
	case 'S':
		/* Kermit transfers can skip files */
		if (q_transfer_stats.protocol == Q_PROTOCOL_KERMIT) {
			kermit_skip_file();
		}
		break;
	case '`':
		/* Backtick works too */
	case KEY_ESCAPE:
		/* ABORT the transfer */
		if ((q_transfer_stats.state != Q_TRANSFER_STATE_END) && (q_transfer_stats.state != Q_TRANSFER_STATE_ABORT)) {
			stop_file_transfer(Q_TRANSFER_STATE_ABORT);
		}

		/* Return to TERMINAL mode */
		switch_state(Q_STATE_CONSOLE);
		return;

	default:
		/* Ignore keystroke */
		break;
	}

} /* ---------------------------------------------------------------------- */

/*
 * protocol_transfer_refresh
 */
void protocol_transfer_refresh() {
	char * status_string;
	int status_left_stop;
	char * message;
	int message_left;
	int window_left;
	int window_top;
	int window_height = 13;
	int window_length = 75;
	int i;
	int percent_complete;
	int batch_percent_complete;
	int cps;

	char time_elapsed_string[SHORT_TIME_SIZE];
	char remaining_time_string[SHORT_TIME_SIZE];
	char batch_time_elapsed_string[SHORT_TIME_SIZE];
	char batch_remaining_time_string[SHORT_TIME_SIZE];
	time_t current_time;
	int hours, minutes, seconds;
	double transfer_time;
	double remaining_time;
	double batch_transfer_time;
	double batch_remaining_time;
	int bits_per_byte = 8;

	/* Compute time */
	/* time_string needs to be hours/minutes/seconds TRANSFER */
	time(&current_time);
	if ((q_transfer_stats.state == Q_TRANSFER_STATE_END) || (q_transfer_stats.state == Q_TRANSFER_STATE_ABORT)) {
		transfer_time = difftime(q_transfer_stats.end_time, q_transfer_stats.file_start_time);
	} else {
		transfer_time = difftime(current_time, q_transfer_stats.file_start_time);
	}
	hours = transfer_time / 3600;
	minutes = ((int)transfer_time % 3600) / 60;
	seconds = (int)transfer_time % 60;
	snprintf(time_elapsed_string, sizeof(time_elapsed_string), "%02u:%02u:%02u", hours, minutes, seconds);

	/* Compute the transfer time and time remaining */
	if (q_transfer_stats.bytes_transfer > 0) {
		remaining_time = (q_transfer_stats.bytes_total - q_transfer_stats.bytes_transfer) * transfer_time / q_transfer_stats.bytes_transfer;
	} else {
		remaining_time = 0;
	}
	if ((q_transfer_stats.state == Q_TRANSFER_STATE_END) || (q_transfer_stats.state == Q_TRANSFER_STATE_FILE_DONE)) {
		remaining_time = 0;
	}

	hours = remaining_time / 3600;
	minutes = ((int)remaining_time % 3600) / 60;
	seconds = (int)remaining_time % 60;
	snprintf(remaining_time_string, sizeof(remaining_time_string), "%02u:%02u:%02u", hours, minutes, seconds);

	/* Batch timings */
	if ((q_transfer_stats.state == Q_TRANSFER_STATE_END) || (q_transfer_stats.state == Q_TRANSFER_STATE_ABORT)) {
		batch_transfer_time = difftime(q_transfer_stats.end_time, q_transfer_stats.batch_start_time);
	} else {
		batch_transfer_time = difftime(current_time, q_transfer_stats.batch_start_time);
	}
	hours = batch_transfer_time / 3600;
	minutes = ((int)batch_transfer_time % 3600) / 60;
	seconds = (int)batch_transfer_time % 60;
	snprintf(batch_time_elapsed_string, sizeof(batch_time_elapsed_string), "%02u:%02u:%02u", hours, minutes, seconds);

	/* Compute the transfer time and time remaining */
	if (q_transfer_stats.bytes_transfer > 0) {
		batch_remaining_time = (q_transfer_stats.batch_bytes_total - q_transfer_stats.batch_bytes_transfer - q_transfer_stats.bytes_transfer) * transfer_time / (q_transfer_stats.batch_bytes_transfer + q_transfer_stats.bytes_transfer);
	} else {
		batch_remaining_time = 0;
	}
	if ((q_transfer_stats.state == Q_TRANSFER_STATE_END) || (q_transfer_stats.state == Q_TRANSFER_STATE_ABORT)) {
		batch_remaining_time = 0;
	}

	/* Let's find the break */
	assert(batch_remaining_time >= 0);

	hours = batch_remaining_time / 3600;
	minutes = ((int)batch_remaining_time % 3600) / 60;
	seconds = (int)batch_remaining_time % 60;
	snprintf(batch_remaining_time_string, sizeof(batch_remaining_time_string), "%02u:%02u:%02u", hours, minutes, seconds);

	/* Filename and pathname could get quite long, let's reduce them. */
	shorten_string(q_transfer_stats.filename, window_length - 10);
	shorten_string(q_transfer_stats.pathname, window_length - 10);

	/* Special case: check the timeout */
	if (((q_transfer_stats.state == Q_TRANSFER_STATE_END) || (q_transfer_stats.state == Q_TRANSFER_STATE_ABORT)) && (q_screen_dirty == Q_FALSE)) {
		transfer_time = difftime(current_time, q_transfer_stats.end_time);

		/* Wait up to 3 seconds */
		if (transfer_time > 3.0) {
			/* Switch back to TERMINAL mode */
			switch_state(Q_STATE_CONSOLE);
			return;
		}
	}

	/*
	 * ASCII special case: we don't put up the dialog window, instead
	 * we update the status line.
	 */
	if (q_transfer_stats.protocol == Q_PROTOCOL_ASCII) {
		/* Max 80 columns wide */
		char status_buffer[80];

		/* Put up the status line */
		screen_put_color_hline_yx(HEIGHT - 1, 0, cp437_chars[HATCH], WIDTH, Q_COLOR_STATUS);

		if (q_program_state == Q_STATE_DOWNLOAD) {
			/* " ASCII DOWNLOAD IN PROGRESS    ESC/`-Save and Exit " */
			status_string = _(" ASCII DOWNLOAD IN PROGRESS    ESC/`-Save and Exit ");

		} else {
			if (q_transfer_stats.bytes_total == 0) {
				/* If the file is 0, percent complete is always 100. */
				percent_complete = 100;
			} else if (q_transfer_stats.bytes_transfer == q_transfer_stats.bytes_total) {
				/* If file is complete percent complete is 100. */
				percent_complete = 100;
			} else {
				/* Else do a true percent complete calculation */
				if (q_transfer_stats.bytes_total > 0) {
					percent_complete = (q_transfer_stats.bytes_transfer * 100) / q_transfer_stats.bytes_total;
				} else {
					percent_complete = 0;
				}
			}

			/* " Uploading %s  Sent = %d    Complete = %d%%  ESC/`-Terminate " */
			sprintf(status_buffer, _(" Uploading %s  Sent = %lu    Complete = %d%%   ESC/`-Terminate "), q_transfer_stats.filename, q_transfer_stats.bytes_transfer, percent_complete);
			status_string = status_buffer;
		}

		status_left_stop = WIDTH - strlen(status_string);
		if (status_left_stop <= 0) {
			status_left_stop = 0;
		} else {
			status_left_stop /= 2;
		}
		screen_put_color_str_yx(HEIGHT - 1, status_left_stop, status_string, Q_COLOR_STATUS);

		screen_flush();
		q_screen_dirty = Q_FALSE;
		return;
	}

	/* Batch upload has more stuff */
	if (q_program_state == Q_STATE_UPLOAD_BATCH) {
		window_height += 3;
	}

	/* Window will be centered on the screen */
	window_left = WIDTH - 1 - window_length;
	if (window_left < 0) {
		window_left = 0;
	} else {
		window_left /= 2;
	}
	window_top = HEIGHT - 1 - window_height;
	if (window_top < 0) {
		window_top = 0;
	} else {
		window_top /= 3;
	}


	if (q_screen_dirty == Q_FALSE) {
		/* Only update the time fields */
		screen_put_color_str_yx(window_top + 6, window_left + 51, _("Time Elapsed "), Q_COLOR_MENU_TEXT);
		screen_put_color_str(time_elapsed_string, Q_COLOR_MENU_COMMAND);
		screen_put_color_str_yx(window_top + 7, window_left + 51, _("++ Remaining "), Q_COLOR_MENU_TEXT);
		screen_put_color_str(remaining_time_string, Q_COLOR_MENU_COMMAND);

		/* Batch time */
		if (q_program_state == Q_STATE_UPLOAD_BATCH) {
			/* Batch times */
			screen_put_color_str_yx(window_top + window_height - 1 - 2, window_left + 2, _("Batch Time Elapsed "), Q_COLOR_MENU_TEXT);
			screen_put_color_str_yx(window_top + window_height - 1 - 2, window_left + 21, batch_time_elapsed_string, Q_COLOR_MENU_COMMAND);
			screen_put_color_str_yx(window_top + window_height - 1 - 2, window_left + 49, _("++ Remaining "), Q_COLOR_MENU_TEXT);
			screen_put_color_str_yx(window_top + window_height - 1 - 2, window_left + 62, batch_remaining_time_string, Q_COLOR_MENU_COMMAND);
		}
		return;
	}

	if (q_program_state == Q_STATE_DOWNLOAD) {
		message = _("Download Status");
		if (q_transfer_stats.protocol == Q_PROTOCOL_KERMIT) {
			status_string = _(" Download in Progress   S-Skip File   ESC/`-Cancel Transfer ");
		} else {
			status_string = _(" Download in Progress   ESC/`-Cancel Transfer ");
		}
	} else if (q_program_state == Q_STATE_UPLOAD) {
		message = _("Upload Status");
		status_string = _(" Upload in Progress   ESC/`-Cancel Transfer ");
	} else {
		message = _("Upload Status");
		if (q_transfer_stats.protocol == Q_PROTOCOL_KERMIT) {
			status_string = _(" Batch Upload in Progress   S-Skip File   ESC/`-Cancel Transfer ");
		} else {
			status_string = _(" Batch Upload in Progress   ESC/`-Cancel Transfer ");
		}
	}

	/* Put up the status line */
	screen_put_color_hline_yx(HEIGHT - 1, 0, cp437_chars[HATCH], WIDTH, Q_COLOR_STATUS);

	status_left_stop = WIDTH - strlen(status_string);
	if (status_left_stop <= 0) {
		status_left_stop = 0;
	} else {
		status_left_stop /= 2;
	}
	screen_put_color_str_yx(HEIGHT - 1, status_left_stop, status_string, Q_COLOR_STATUS);

	/* Draw the sub-window */
	screen_draw_box(window_left, window_top, window_left + window_length, window_top + window_height);

	/* Place the title */
	message_left = window_length - (strlen(message) + 2);
	if (message_left < 0) {
		message_left = 0;
	} else {
		message_left /= 2;
	}
	screen_put_color_printf_yx(window_top + 0, window_left + message_left, Q_COLOR_WINDOW_BORDER, " %s ", message);

	/* Protocol name, filename, pathname */
	screen_put_color_str_yx(window_top + 3, window_left + 2, _("File "), Q_COLOR_MENU_TEXT);
	screen_put_color_str(q_transfer_stats.filename, Q_COLOR_MENU_COMMAND);
	screen_put_color_str_yx(window_top + 1, window_left + 27, _("Protocol "), Q_COLOR_MENU_TEXT);
	screen_put_color_str(q_transfer_stats.protocol_name, Q_COLOR_MENU_COMMAND);

	screen_put_color_str_yx(window_top + 4, window_left + 2, _("Path "), Q_COLOR_MENU_TEXT);
	screen_put_color_str(q_transfer_stats.pathname, Q_COLOR_MENU_COMMAND);

	/* Bytes and blocks total fields */
	screen_put_color_str_yx(window_top + 6, window_left + 2, _("Bytes Total "), Q_COLOR_MENU_TEXT);
	screen_put_color_printf(Q_COLOR_MENU_COMMAND, "%-lu", q_transfer_stats.bytes_total);
	screen_put_color_str_yx(window_top + 6, window_left + 27, _("Blocks Total "), Q_COLOR_MENU_TEXT);
	screen_put_color_printf(Q_COLOR_MENU_COMMAND, "%-lu", q_transfer_stats.blocks);

	/* Time fields */
	screen_put_color_str_yx(window_top + 6, window_left + 51, _("Time Elapsed "), Q_COLOR_MENU_TEXT);
	screen_put_color_str(time_elapsed_string, Q_COLOR_MENU_COMMAND);
	screen_put_color_str_yx(window_top + 7, window_left + 51, _("++ Remaining "), Q_COLOR_MENU_TEXT);
	screen_put_color_str(remaining_time_string, Q_COLOR_MENU_COMMAND);

	/* Bytes and blocks transferred fields */
	if (q_program_state == Q_STATE_DOWNLOAD) {
		screen_put_color_str_yx(window_top + 7, window_left + 2, _("Bytes Rcvd  "), Q_COLOR_MENU_TEXT);
	} else {
		screen_put_color_str_yx(window_top + 7, window_left + 2, _("Bytes Sent  "), Q_COLOR_MENU_TEXT);
	}
	screen_put_color_printf(Q_COLOR_MENU_COMMAND, "%-lu", q_transfer_stats.bytes_transfer);
	if (q_program_state == Q_STATE_DOWNLOAD) {
		screen_put_color_str_yx(window_top + 7, window_left + 27, _("Blocks Rcvd  "), Q_COLOR_MENU_TEXT);
	} else {
		screen_put_color_str_yx(window_top + 7, window_left + 27, _("Blocks Sent  "), Q_COLOR_MENU_TEXT);
	}
	screen_put_color_printf(Q_COLOR_MENU_COMMAND, "%-lu", q_transfer_stats.blocks_transfer);

	/* Block size, error count, and efficiency */
	screen_put_color_str_yx(window_top + 8, window_left + 2, _("Error Count "), Q_COLOR_MENU_TEXT);
	screen_put_color_printf(Q_COLOR_MENU_COMMAND, "%-lu", q_transfer_stats.error_count);
	screen_put_color_str_yx(window_top + 8, window_left + 27, _("Block Size   "), Q_COLOR_MENU_TEXT);
	screen_put_color_printf(Q_COLOR_MENU_COMMAND, "%-lu", q_transfer_stats.block_size);

	/* CPS */
	screen_put_color_str_yx(window_top + 9, window_left + 51, _("Chars/second "), Q_COLOR_MENU_TEXT);
	cps = q_transfer_stats.bytes_transfer / transfer_time;
	if (cps > q_transfer_stats.bytes_transfer) {
		cps = q_transfer_stats.bytes_transfer;
	}
	screen_put_color_printf(Q_COLOR_MENU_COMMAND, "%-lu", cps);

	screen_put_color_str_yx(window_top + 8, window_left + 51, _("Efficiency   "), Q_COLOR_MENU_TEXT);
	if (q_status.serial_open == Q_TRUE) {
		switch (q_serial_port.data_bits) {
		case DATA_BITS_8:
			bits_per_byte = 8;
			break;
		case DATA_BITS_7:
			bits_per_byte = 7;
			break;
		case DATA_BITS_6:
			bits_per_byte = 6;
			break;
		case DATA_BITS_5:
			bits_per_byte = 5;
			break;
		}
		switch (q_serial_port.stop_bits) {
		case STOP_BITS_1:
			bits_per_byte += 1;
			break;
		case STOP_BITS_2:
			bits_per_byte += 2;
			break;
		}

		/* Add mark bit */
		bits_per_byte += 1;

		if (q_serial_port.dce_baud > 0) {
			screen_put_color_printf(Q_COLOR_MENU_COMMAND, " %6.2f%%", ((float)cps * bits_per_byte * 100.0) / (float)q_serial_port.dce_baud);
		} else {
			screen_put_color_str("N/A", Q_COLOR_MENU_COMMAND);
		}
	} else {
		screen_put_color_str("N/A", Q_COLOR_MENU_COMMAND);
	}


	/* Last message */
	screen_put_color_str_yx(window_top + 10, window_left + 2, _("Status Msgs "), Q_COLOR_MENU_TEXT);
	screen_put_color_str(q_transfer_stats.last_message, Q_COLOR_MENU_COMMAND);

	/* Three cases:
	 * 1) Transfer complete
	 * 2) Transfer aborted
	 * 3) Transfer in progress
	 */
	if ((q_transfer_stats.state == Q_TRANSFER_STATE_END) || (q_transfer_stats.state == Q_TRANSFER_STATE_FILE_DONE)) {
		/* File complete, percent complete is 100. */
		percent_complete = 100;
	} else if (q_transfer_stats.bytes_transfer == 0) {
		/* 0 bytes transferred, percent complete is 0. */
		percent_complete = 0;
	} else if ((q_program_state == Q_STATE_DOWNLOAD) && ((q_transfer_stats.protocol == Q_PROTOCOL_XMODEM) ||
		   (q_transfer_stats.protocol == Q_PROTOCOL_XMODEM_RELAXED) ||
		   (q_transfer_stats.protocol == Q_PROTOCOL_XMODEM_CRC) ||
		   (q_transfer_stats.protocol == Q_PROTOCOL_XMODEM_1K) ||
		   (q_transfer_stats.protocol == Q_PROTOCOL_XMODEM_1K_G)
	)) {
		/* Xmodem does not report file size on downloads, percent complete is always 0 */
		percent_complete = 0;
	} else if (q_transfer_stats.bytes_total == 0) {
		/* 0-byte files have percent complete = 0 */
		percent_complete = 0;
	} else if (q_transfer_stats.bytes_transfer == q_transfer_stats.bytes_total) {
		/* Another check: if file is complete percent complete is 100. */
		percent_complete = 100;
	} else {
		/* Finally, fall through to a true percent complete calculation */
		percent_complete = (q_transfer_stats.bytes_transfer * 100) / q_transfer_stats.bytes_total;
	}
	if (percent_complete > 100) {
		percent_complete = 100;
	}

	screen_put_color_str_yx(window_top + 11, window_left + 2, _("Completion  "), Q_COLOR_MENU_TEXT);
	screen_move_yx(window_top + 11, window_left + 14);
	screen_put_color_printf(Q_COLOR_MENU_COMMAND, "%-3u%%   ", percent_complete);
	for (i=0; i<(percent_complete * 50) / 100; i++) {
		screen_put_color_char(cp437_chars[HATCH], Q_COLOR_MENU_COMMAND);
	}
	for (; i<50; i++) {
		screen_put_color_char(cp437_chars[BOX], Q_COLOR_MENU_COMMAND);
	}

	/* Batch upload */
	if (q_program_state == Q_STATE_UPLOAD_BATCH) {
		if (q_transfer_stats.batch_bytes_transfer + q_transfer_stats.bytes_transfer == 0) {
			batch_percent_complete = 0;
		} else if (q_transfer_stats.batch_bytes_transfer + q_transfer_stats.bytes_transfer >= q_transfer_stats.batch_bytes_total) {
			batch_percent_complete = 100;
		} else if (q_transfer_stats.batch_bytes_total == 0) {
			/* If we ever send a batch of 0-byte files... */
			batch_percent_complete = 0;
		} else {
			batch_percent_complete = ((q_transfer_stats.batch_bytes_transfer + q_transfer_stats.bytes_transfer) * 100) / q_transfer_stats.batch_bytes_total;
		}
		if (batch_percent_complete > 100) {
			batch_percent_complete = 100;
		}

		message = _("Batch Upload Status");

		screen_put_color_hline_yx(window_top + window_height - 1 - 3, window_left + 1, cp437_chars[Q_WINDOW_TOP], window_length - 2, Q_COLOR_WINDOW_BORDER);
		screen_put_color_char_yx(window_top + window_height - 1 - 3, window_left + 0, cp437_chars[Q_WINDOW_LEFT_TEE], Q_COLOR_WINDOW_BORDER);
		screen_put_color_char_yx(window_top + window_height - 1 - 3, window_left + window_length - 1, cp437_chars[Q_WINDOW_RIGHT_TEE], Q_COLOR_WINDOW_BORDER);

		message_left = window_length - (strlen(message) + 2);
		if (message_left < 0) {
			message_left = 0;
		} else {
			message_left /= 2;
		}
		screen_put_color_printf_yx(window_top + window_height - 1 - 3, window_left + message_left, Q_COLOR_WINDOW_BORDER, " %s ", message);

		/* Batch times */
		screen_put_color_str_yx(window_top + window_height - 1 - 2, window_left + 2, _("Batch Time Elapsed "), Q_COLOR_MENU_TEXT);
		screen_put_color_str_yx(window_top + window_height - 1 - 2, window_left + 21, batch_time_elapsed_string, Q_COLOR_MENU_COMMAND);
		screen_put_color_str_yx(window_top + window_height - 1 - 2, window_left + 49, _("++ Remaining "), Q_COLOR_MENU_TEXT);
		screen_put_color_str_yx(window_top + window_height - 1 - 2, window_left + 62, batch_remaining_time_string, Q_COLOR_MENU_COMMAND);

		/* Progress bar */
		screen_put_color_str_yx(window_top + window_height - 1 - 1, window_left + 2, _("Completion  "), Q_COLOR_MENU_TEXT);
		screen_move_yx(window_top + window_height - 1 - 1, window_left + 14);
		screen_put_color_printf(Q_COLOR_MENU_COMMAND, "%-3u%%   ", batch_percent_complete);
		for (i=0; i<(batch_percent_complete * 50) / 100; i++) {
			screen_put_color_char(cp437_chars[HATCH], Q_COLOR_MENU_COMMAND);
		}
		for (; i<50; i++) {
			screen_put_color_char(cp437_chars[BOX], Q_COLOR_MENU_COMMAND);
		}
	}

	screen_flush();
	q_screen_dirty = Q_FALSE;
} /* ---------------------------------------------------------------------- */
